package com.platomix.api.cgi.controller;

import cn.hutool.core.io.IoUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.http.ContentType;
import cn.hutool.http.HttpRequest;
import cn.hutool.http.HttpResponse;
import cn.hutool.http.HttpUtil;
import cn.hutool.http.Method;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.platomix.api.cgi.entity.Api;
import com.platomix.api.cgi.entity.ApiLog;
import com.platomix.api.cgi.entity.App;
import com.platomix.api.cgi.entity.AppApi;
import com.platomix.api.cgi.service.ApiLogService;
import com.platomix.api.cgi.service.ApiService;
import com.platomix.api.cgi.service.AppApiService;
import com.platomix.api.cgi.service.AppService;
import com.platomix.api.cgi.support.properties.HttpTimeoutProperties;
import lombok.AllArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.time.Duration;
import java.util.*;

@Slf4j
@RestController
@RequestMapping("/")
@AllArgsConstructor
public class CgiController {

    private final ApiService apiService;
    private final ApiLogService apiLogService;
    private final AppService appService;
    private final AppApiService appApiService;
    private final StringRedisTemplate stringRedisTemplate;
    private final HttpTimeoutProperties httpTimeoutProperties;

    //cgi/a461f3f77e774105b1f02abe5aec6ba8/67
    @RequestMapping("cgi/{uuid}/{id}")
    public String liveness(@PathVariable("uuid") String uuid, @PathVariable("id") Integer id, HttpServletRequest request, HttpServletResponse response) throws Exception {
        String token = request.getHeader("access_token");
        if (StrUtil.isBlank(token)) {
            return "无效的token";
        }
        String appIdStr = stringRedisTemplate.opsForValue().get(token);
        if (StrUtil.isBlank(appIdStr)) {
            return "token不存在";
        }
        ApiLog apiLog = new ApiLog();
        apiLog.setInvokeDate(new Date());
        Api api = apiService.getById(id);
        String remoteIP = getRemoteIP(request);
        String whiteList = api.getWhiteList();
        String blackList = api.getBlackList();
        if (StrUtil.isNotEmpty(whiteList) && !"null".equals(whiteList)) {
            List<String> list = Arrays.asList(whiteList.split(","));
            if (!list.contains(remoteIP)) {
                return "请求ip不在白名单内";
            }
        }
        if (StrUtil.isNotEmpty(blackList) && !"null".equals(blackList)) {
            List<String> list = Arrays.asList(blackList.split(","));
            if (list.contains(remoteIP)) {
                return "请求ip在黑名单内";
            }
        }
        apiLog.setApiId(api.getApiId());

        App app = appService.getById(Integer.parseInt(appIdStr));
        if (app == null) {
            return "应用不存在";
        }
        if (app.getStatus() != 1) {
            return "应用已停用";
        }
        AppApi appApi = appApiService.getOne(Wrappers.<AppApi>lambdaQuery().eq(AppApi::getAppId, app.getAppId()).eq(AppApi::getApiId, api.getApiId()).last("limit 1"));
        if (appApi == null) {
            return "接口信息不存在";
        }
        if (appApi.getEnable() == 1) {
            return "接口已停用";
        }
        String key = token + ":" + app.getAppId() + ":" + appApi.getApiId();
        Long increment = stringRedisTemplate.opsForValue().increment(key);
        //每秒的第一次，设置过期时间1s
        if (increment == 1) {
            stringRedisTemplate.expire(key, Duration.ofSeconds(1));
        }
        if (appApi.getQpsLimit() != null && increment.intValue() > Integer.valueOf(appApi.getQpsLimit())) {
            return "该接口已超过qps限制";
        }
        apiLog.setAppId(Integer.parseInt(appIdStr));
        apiLog.setUserId(app.getUserId());
        //构造请求方法
        Method method = Method.GET;
        try {
            Method.valueOf(request.getMethod().toUpperCase());
        } catch (Exception e) {
            log.warn("未找到请求method[{}],使用默认值:GET", request.getMethod());
        }
        String result = "接口请求失败";
        try {
            result = this.httpInvoke(apiLog, request, api.getUrl(), method);
        } catch (Exception e) {
            log.error(e.getMessage(), e);
        }
        apiLogService.save(apiLog);
        return result;
    }

    /**
     * <p>Http请求</p>
     *
     * @param apiLog:      日志对象
     * @param request:     请求对象
     * @param redirectUrl: 请求地址
     * @param method:      方法
     * @return java.lang.String
     * @author wookvn
     * @datetime 2024-11-22 16:29:41
     **/
    private String httpInvoke(ApiLog apiLog, HttpServletRequest request, String redirectUrl, Method method) throws IOException {
        String params = request.getQueryString();
        if (params != null && !params.trim().isEmpty()) {
            redirectUrl = redirectUrl + "?" + params;
            apiLog.setParams(params);
        }
        Map<String, String> headerMap = new HashMap<>();
        Enumeration<String> headerNames = request.getHeaderNames();
        while (headerNames.hasMoreElements()) {
            String headerName = headerNames.nextElement();
            String headerValue = request.getHeader(headerName);
            log.debug("headerName={}, headerValue={}", headerName, headerValue);
            if (StrUtil.containsAnyIgnoreCase(headerName, "host", "content-length")) {
                continue;
            }
            headerMap.put(headerName, headerValue);
        }
        apiLog.setInvokeTime(System.currentTimeMillis() + "");
        //使用hutool的httpUtil构造请求
        HttpRequest httpRequest = HttpUtil.createRequest(method, redirectUrl)
                .addHeaders(headerMap)
                .setConnectionTimeout(httpTimeoutProperties.getConnect())
                .setReadTimeout(httpTimeoutProperties.getRead())
                .setReadTimeout(httpTimeoutProperties.getRead());
        //POST请求的时候，设置JSON参数
        if (method == Method.POST) {
            String jsonParams = IoUtil.readUtf8(request.getInputStream());
            httpRequest.body(jsonParams, ContentType.JSON.getValue());
        }
        //开始请求
        HttpResponse httpResponse = httpRequest.execute();
        String result = httpResponse.body();
        apiLog.setResponseTime(System.currentTimeMillis() + "");
        log.debug("[{}]HttpResult:{}", method, result);
        apiLog.setErrCode(httpResponse.getStatus() + "");
        if (httpResponse.getStatus() == 200) {
            apiLog.setCallResult(1);
        } else {
            apiLog.setCallResult(0);
        }
        return result;
    }

    private String getRemoteIP(HttpServletRequest request) {
        String ip = request.getHeader("x-forwarded-for");
        if (ip == null || ip.length() == 0 || "unknown".equalsIgnoreCase(ip)) {
            ip = request.getHeader("X-Real-IP");
        }
        if (ip == null || ip.length() == 0 || "unknown".equalsIgnoreCase(ip)) {
            ip = request.getHeader("WL-Proxy-Client-IP");
        }
        if (ip == null || ip.length() == 0 || "unknown".equalsIgnoreCase(ip)) {
            ip = request.getRemoteAddr();
        }
        return ip;
    }

}
